import os
import time
import tempfile
import shutil
import xml.etree.ElementTree as ET
from selenium import webdriver
from gmapsbounds import utils
from gmapsbounds import reader
from gmapsbounds import constants
from gmapsbounds import llpx

class Location:
    def __init__(self, driver, name):
        self.driver = driver
        self.name = name
        self.save_dir = tempfile.mkdtemp()
        self.center = ''
        self.zoom = ''
        self.polygons = []

    def download_center_map(self):
        self.load_map()
        self.center, self.zoom = utils.parse_url(self.driver.current_url)
        center_list = [float(coord) for coord in self.center.split(', ')]
        if self.center in constants.ZOOM_OUT:
            self.zoom_out()
        self.offset = llpx.lat_lng_to_pixels(center_list[0], center_list[1], self.zoom)
        self.take_map_screenshot('boundaries.png')

    def zoom_out(self):
        self.zoom = str(int(self.zoom) - 1)
        current_url = self.driver.current_url
        self.driver.get('https://www.google.com/maps/place/{}/@{},{}z/'.format(
        self.name.replace(' ', '+'), self.center.replace(' ', ''), self.zoom))
        while current_url == self.driver.current_url:
            pass

    def load_map(self):
        self.driver.get('https://www.google.com/maps/place/{}'.format(self.name.replace(' ', '+')))
        while (['https:', '', 'www.google.com', 'maps', 'place'] !=
            self.driver.current_url.split('/')[:5] or '@' not in self.driver.current_url):
            pass

    def take_map_screenshot(self, filename):
        container = self.driver.find_element_by_class_name('widget-scale')
        # clicking this closes the ad panel
        container.click()
        time.sleep(1)
        self.driver.save_screenshot(os.path.join(self.save_dir, filename))
        shutil.copyfile(os.path.join(self.save_dir, filename), filename)

    def calculate_coordinates(self):
        boundaries_filename = os.path.join(self.save_dir, 'boundaries.png')
        rgb_im = utils.load_image(boundaries_filename)
        nodes = reader.get_nodes(rgb_im)
        self.polygons = reader.get_polygons(nodes, rgb_im)
        self.polygons = reader.prune_nodes(self.polygons)
        self.polygons = reader.trim_overlapping_nodes(self.polygons)
        self.pixcenter = [rgb_im.size[0] // 2, rgb_im.size[1] // 2]
        for poly in self.polygons:
            for node in poly.nodes:
                node.location = self

    def write_to_kml(self, filename):
        root = ET.Element('kml')
        root.set('xmlns', 'http://www.opengis.net/kml/2.2')
        placemark = ET.SubElement(root, 'Placemark')
        placemark.append(utils.makeelement('name', self.name))
        for poly in self.polygons:
            polygon = ET.SubElement(placemark, 'Polygon')
            polygon.append(utils.makeelement('extrude', '1'))
            polygon.append(utils.makeelement('altitudeMode', 'relativeToGround'))
            boundary = ET.SubElement(polygon, 'outerBoundaryIs')
            linear_ring = ET.SubElement(boundary, 'LinearRing')
            coordinates = ET.SubElement(linear_ring, 'coordinates')
            text_list = []
            for node in poly.nodes:
                coords_as_strings = [str(coord) for coord in node.get_lat_lng()]
                text_list.append(','.join(reversed(coords_as_strings)))
            coordinates.text = '\n'.join(text_list)
        with open(filename, 'w') as f:
            f.write(ET.tostring(root).decode('utf8'))




from gmapsbounds import utils
from gmapsbounds import constants
from gmapsbounds import llpx
from gmapsbounds import polygon

def get_nodes(rgb_image):
    nodes = []
    width, height = rgb_image.size
    do_not_check = set()
    menu_borders = utils.get_menu_borders(rgb_image)
    for x in range(width):
        for y in range(height):
            if ((y in range(menu_borders[0][0], menu_borders[0][1] + 1) and
                x in range (menu_borders[1][0], menu_borders[1][1] + 1)) or
                (x, y) in do_not_check):
                continue
            r, g, b = rgb_image.getpixel((x, y))
            if [r, g, b] == constants.INVALID[2]:
                exclude_surrounding_nodes(x, y, do_not_check, rgb_image)
            if valid_color(r, g, b) and no_invalid_adjacent(x, y, rgb_image):
                nodes.append(Node(x, y))
    nodes[0].visited = True
    return nodes

def prune_nodes(nodes):
    pruned_polygons = []
    for poly in nodes:
        assert len(poly.nodes) > 2
        pruned = polygon.Polygon(poly.nodes[:2])
        for node in poly.nodes[2:]:
            if (utils.get_slope(pruned.nodes[-2], pruned.nodes[-1]) == utils.get_slope(pruned.nodes[-2], node) or
                utils.get_distance(pruned.nodes[-1], node)) <= 2:
                pruned.nodes.pop()
            pruned.nodes.append(node)
        if len(pruned.nodes) > 2:
            pruned_polygons.append(pruned)
    return pruned_polygons

def exclude_surrounding_nodes(x, y, nodes_to_exclude, rgb_im, depth=5):
    for i in range(-depth, depth + 1):
        for j in range(-depth, depth + 1):
            r, g, b = rgb_im.getpixel((x+i, y+j))
            if [r, g, b] != constants.INVALID[2] or (i == 0 and j == 0):
                try:
                    nodes_to_exclude.add((x+i, y+j))
                except:
                    pass

def valid_color(r, g, b):
    if ([r, g, b] in constants.VALID or
        (r in constants.ALLOWABLE_RED and g in constants.ALLOWABLE_GREEN and
        b in constants.ALLOWABLE_BLUE and abs(g - b) < constants.MAX_BLUE_GREEN_DIFFERENCE)):
        return True
    return False

def no_invalid_adjacent(x, y, image):
    for i in range(-2, 3):
        for j in range(-2, 3):
            try:
                r, g, b = image.getpixel((x + i, y + j))
                if [r, g, b] in constants.INVALID or (r in constants.ALLOWABLE_RED and 100 > g == b):
                    return False
            except IndexError:
                return False
    return True

def get_polygons(nodes, rgb_im):
    polygons = []
    unvisited = [node for node in nodes if node.visited is False]
    while unvisited:
        poly = polygon.Polygon()
        current = unvisited[0]
        current.visited = True
        closest, distance = get_closest_unvisited_node(current, nodes, rgb_im)
        while closest is not None:
            poly.nodes.append(current)
            current = closest
            current.visited = True
            closest, distance = get_closest_unvisited_node(current, nodes, rgb_im)
            if closest is None:
                break
            i = -1
            while distance > constants.MAX_NODE_DIFFERENCE:
                if current is poly.nodes[0] or i < -constants.MAX_NODE_BACKTRACK:
                    closest = None
                    break
                current = poly.nodes[i]
                closest, distance = get_closest_unvisited_node(current, unvisited, rgb_im)
                i -= 1
        if len(poly.nodes) > 2:
            polygons.append(poly)
        unvisited = [node for node in nodes if node.visited is False]
    return polygons

def trim_overlapping_nodes(polygons):
    assert polygons
    polygons = utils.sort_by_polygon_length(polygons)
    polygons.reverse()
    exterior_polygons = [polygons[0]]
    for test_polygon in polygons[1:]:
        starting_count = len(test_polygon.nodes)
        for exterior_polygon in exterior_polygons:
            exterior_nodes = test_polygon.get_exterior_nodes(exterior_polygon)
            if not exterior_nodes:
                if len(test_polygon.nodes) == starting_count:
                    exterior_polygon.inner = test_polygon
                break
            if exterior_polygon is exterior_polygons[-1] and len(exterior_nodes) > 2:
                test_polygon.nodes = exterior_nodes
                exterior_polygons.append(test_polygon)
                break
    return exterior_polygons


def remove_overlapping_polygons(polygons):
    assert polygons
    polygons = utils.sort_by_polygon_length(polygons)
    polygons.reverse()
    okay_polygons = [polygons[0]]
    for test_polygon in polygons:
        overlap = False
        if test_polygon in okay_polygons:
            continue
        for okay_polygon in okay_polygons:
            if test_polygon.overlaps(okay_polygon):
                overlap = True
                break
        if not overlap:
            okay_polygons.append(test_polygon)
    return okay_polygons

def get_closest_unvisited_node(current, nodes, rgb_im):
    closest_node = None
    shortest_distance = None
    pos = nodes.index(current)
    i = 1
    go_up = True
    go_down = True
    while (0 <= pos - i or len(nodes) > pos + i) and (go_up or go_down):
        signs = []
        if go_down:
            signs.append(-1)
        if go_up:
            signs.append(1)
        for sign in signs:
            index = pos + i*sign
            if not 0 <= index < len(nodes):
                continue
            node = nodes[index]
            if closest_node is not None:
                if sign == -1 and shortest_distance < current.x - node.x:
                    go_down = False
                elif sign == 1 and shortest_distance < node.x - current.x:
                    go_up = False
            if node.visited:
                continue
            distance = utils.get_distance(nodes[pos], node)
            distance *= utils.get_water_multiplier(current, node, rgb_im)
            if shortest_distance is None or distance < shortest_distance:
                closest_node = node
                shortest_distance = distance
        i += 1
    return closest_node, shortest_distance

def add_skipped_nodes(x, y, do_not_check):
    do_not_check.add((x+1, y-1))
    do_not_check.add((x+1, y))
    do_not_check.add((x+1, y+1))
    do_not_check.add((x, y+1))

class Node:
    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.location = None
        self.visited = False

    def get_lat_lng(self):
        return llpx.pixels_to_lat_lng(self.location.offset[0] - self.location.pixcenter[0] + self.x,
            self.location.offset[1] - self.location.pixcenter[1] + self.y, self.location.zoom)

    def __str__(self):
        return '<Node at {}, {}>'.format(self.x, self.y)

    def __repr__(self):
        return self.__str__()


import os
import re
import time
from PIL import Image
import xml.etree.ElementTree as ET
import xvfbwrapper
from selenium import webdriver
from gmapsbounds import constants

def load_image(filename):
    im = Image.open(filename)
    im.load()
    return im.convert('RGB')

def parse_url(url):
    spl = url.split('/')
    assert ['https:', '', 'www.google.com', 'maps', 'place'] == spl[:5]
    lat_lng = re.search(r'@([-0-9.]+,[-0-9.]+)', spl[6]).group(1)
    lat_lng = lat_lng.replace(',', ', ')
    zoom = re.search(r'(\d{1,2})z$', spl[6]).group(1)
    return lat_lng, zoom

def launch_driver(url=''):
    driver = webdriver.Firefox()
    driver.maximize_window()
    if url:
        driver.get(url)
    return driver

def wait_until(func, arg, interval=1):
    while True:
        try:
            return func(arg)
        except:
            time.sleep(interval)

def ray_crosses_edge(point, v1, v2):
    '''
    Test whether a point lies within a polygon, excluding vertices and edges
    '''
    # All three args are nodes with x and y properties
    # Ray goes north for this function
    if (point.x > max(v1.x, v2.x) or point.x < min(v1.x, v2.x) or
        point.y < min(v1.y, v2.y) or (v1.x == point.x and v1.x <= v2.x)):
        return False
    return not (point.y <= v2.y - (v2.x - point.x) * get_slope(v1, v2))

def max_length_list(alist):
    max_list = None
    for sublist in alist:
        if max_list is None or len(sublist) > len(max_list):
            max_list = sublist
    return max_list

def get_water_multiplier(node1, node2, rgb_image):
    '''
    Return approximate water multiplication factor between two nodes
    '''
    x = node1.x
    y = node1.y
    count = 0
    iterations = 10
    for i in range(iterations):
        x += (node2.x - node1.x) / iterations
        y += (node2.y - node1.y) / iterations
        r, g, b = rgb_image.getpixel((x, y))
        if is_water(r, g, b):
            count += 1
    return 1 + count * .5

def is_water(r, g, b):
    if r in constants.WATER_RED and g in constants.WATER_GREEN and b in constants.WATER_BLUE:
        return True
    return False

def get_slope(node1, node2):
    try:
        return (node2.y-node1.y) / (node2.x - node1.x)
    except:
        return None

def get_distance(node1, node2):
    return((node1.x - node2.x)**2 + (node1.y - node2.y)**2)**.5

def get_midpoint(node1, node2):
    return (node1.x + (node2.x - node1.x) / 2, node1.y + (node2.y - node1.y) / 2)

def makeelement(tag, text=None, attributes=None, tail=None):
    if attributes is None:
        attributes = {}
    newelement = ET.Element(tag)
    for k, v in attributes.items():
        newelement.set(k, v)
    if text is not None: newelement.text = text
    if tail is not None: newelement.tail = tail
    return newelement


def sort_by_polygon_length(polygons):
    less = []
    equal = []
    greater = []
    if len(polygons) > 1:
        pivot = polygons[0]
        for poly in polygons:
            if len(poly.nodes) < len(pivot.nodes):
                less.append(poly)
            if len(poly.nodes) == len(pivot.nodes):
                equal.append(poly)
            if len(poly.nodes) > len(pivot.nodes):
                greater.append(poly)
        return sort_by_polygon_length(less) + equal + sort_by_polygon_length(greater)
    return polygons

def get_menu_borders(rgb_image):
    for x in range(30, 430):
        r, g, b = rgb_image.getpixel((x, 370))
        if [r, g, b] != [255, 255, 255]:
            return constants.MENU_BORDERS
    return constants.MENU_BORDERS_WITH_ADS




USER_AGENT = (
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/36.0.1985.143 Safari/537.36"
)

ZOOM_OUT = {
	'37.7577, -122.4376': 'San Francisco, CA'
}

MAX_NODE_DIFFERENCE = 75
MAX_NODE_BACKTRACK = 50

#Ny, Sy, Wx, Ex
MENU_BORDERS = [[15, 123], [30, 429]]
MENU_BORDERS_WITH_ADS = [[15, 447], [30, 429]]

WATER_RED = list(range(148, 185))
WATER_GREEN = list(range(193, 215))
WATER_BLUE = list(range(253, 256))

ALLOWABLE_RED = list(range(218, 241))
ALLOWABLE_GREEN = list(range(0, 191))
ALLOWABLE_BLUE = list(range(65, 181))

MAX_BLUE_GREEN_DIFFERENCE = 20

VALID = [[232, 151, 142], [237, 162, 155]]
INVALID = [[215, 60, 48], [218, 97, 97], [179, 56, 44], [180, 56, 44]]
# Highway signs (domestic), Hospitals, Highway signs (foreign), city names


from gmapsbounds import utils

class Polygon:
    def __init__(self, nodes=None):
        if nodes is None:
            nodes = []
        self.nodes = nodes
        self.inner = None

    def contains(self, point):
        assert len(self.nodes) > 2
        crossings = 0
        i = -len(self.nodes) + 1
        for node in self.nodes:
            if utils.ray_crosses_edge(point, node, self.nodes[i]):
                crossings += 1
            i += 1
        return crossings % 2 == 1

    def get_exterior_nodes(self, poly):
        assert len(self.nodes) > 2
        exterior_nodes = []
        replace_list = []
        for node in self.nodes:
            if not poly.contains(node):
                replace_list.append(node)
            else:
                if len(replace_list) > len(exterior_nodes):
                    exterior_nodes = replace_list
                replace_list = []
        if len(replace_list) > len(exterior_nodes):
            exterior_nodes = replace_list
        return exterior_nodes


from setuptools import setup, find_packages
from codecs import open
from os import path
here = path.abspath(path.dirname(__file__))

setup(
	name='gmapsbounds',
	version='0.1.3',
	description='Extract Lat/Lng boundary points of geographical regions from Google Maps',
	url='https://github.com/evfredericksen/gmapsbounds',
	author='Evan Fredericksen',
	author_email='evfredericksen@gmail.com',
	license='MIT',
	classifiers=[
	'Development Status :: 3 - Alpha',
	'Intended Audience :: Developers',
	'Topic :: Software Development :: Build Tools',
	'License :: OSI Approved :: MIT License',
	'Programming Language :: Python :: 2',
	'Programming Language :: Python :: 2.6',
	'Programming Language :: Python :: 2.7',
	'Programming Language :: Python :: 3',
	'Programming Language :: Python :: 3.2',
	'Programming Language :: Python :: 3.3',
	'Programming Language :: Python :: 3.4',
	],
	keywords='google maps lat lng boundaries',
	packages=['gmapsbounds'],
	package_dir = {
		'gmapsbounds': 'gmapsbounds',
	},
	install_requires=['selenium'],
	scripts = ['scripts/gmapsbounds'],
	include_package_data = True,
	long_description = '''\
	Use Google Maps web app to find lat/lng boundaries of
	geographical regions, including cities, ZIP codes, counties,
	states and countries.
	'''
)